--------------------------------------------------------------------------------
-- Company: FSU-FAMU
-- Engineer: Dr. Uwe Meyer-Baese
-- Create Date:    10:56:41 01/01/11
-- Design Name:    i2c_config
-- Project Name:    DE2_lab
-- Target Device:   EP2C35F672C6
-- Tool versions:   Quartus II 9.1
-- Description:   This is the parallel to serial converter for the CODEC data.
--                The output sampling rate is 48 kHz; LR control clock is the 
--                same for ADC and DAC.
--                The controller runs with 384*48 kHz = 18.432 MHz
--                It also generates the left/right channel control signal "lrck."
--
-- Additional Comments:  Resources used:
--  Total Number LEs: 34         
--  Number of 9x9 Mults: 0
--  Number of Embedded 4Ks: 0
--  Maximum Delay is 2.54 ns => 393.70 MHz
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--------------------------------------------------------------------------------
ENTITY audio_dac IS
  PORT( --  Control Signals incomming
     irst_n    : IN STD_LOGIC; -- Input reset
     iclk_18_4 : IN STD_LOGIC; -- Input 18 MHz clock from PLL
     idata16   : IN STD_LOGIC_VECTOR(15 DOWNTO 0); -- 16 bit data in
     --  Audio Side outgoing
     oaud_data : OUT STD_LOGIC; -- Single output bit stream
     oaud_lrck : OUT STD_LOGIC; -- Left =1  and right=0 channel clock
     oaud_bck  : BUFFER STD_LOGIC); -- Bit clock for output
END ENTITY audio_dac; 
      
ARCHITECTURE fpga OF audio_dac IS
  
  CONSTANT  REF_CLK     : INTEGER :=  18432000;  --  18.432 MHz
  CONSTANT  SAMPLE_RATE : INTEGER :=  48000;     --  48 KHz
  CONSTANT  DATA_WIDTH  : INTEGER :=  16;        --  16 Bits
  CONSTANT  CHANNEL_NUM : INTEGER :=  2;         --  Dual channel

--  Internal Registers and Wires
  SIGNAL bck_div : STD_LOGIC_VECTOR(3 DOWNTO 0);
  SIGNAL lrck_1x_div : STD_LOGIC_VECTOR(8 DOWNTO 0);
  SIGNAL sel_cont : STD_LOGIC_VECTOR(3 DOWNTO 0);
  SIGNAL nsel_cont : STD_LOGIC_VECTOR(4 DOWNTO 0);
  SIGNAL lrck_1x : STD_LOGIC;
  
BEGIN
--------------------------------------------------  
------------  output aud_bck generator  --------------
  P1: PROCESS
  BEGIN
    WAIT UNTIL iclk_18_4 = '1';
    IF irst_n = '0' THEN
      bck_div    <=  (OTHERS => '0');
      oaud_bck   <=  '0';
    ELSE          -- about 6 clock cycles low/high per bit
      IF bck_div >= REF_CLK/(SAMPLE_RATE*DATA_WIDTH*CHANNEL_NUM*2)-1 THEN
        bck_div    <=  (OTHERS => '0');
        oaud_bck   <=  NOT oaud_bck;
      ELSE
        bck_div    <=  bck_div + 1;
      END IF;
    END IF;
  END PROCESS;
  
--------------------------------------------------
------------  AUD_LRCK Generator  --------------
  P2: PROCESS
  BEGIN
    WAIT UNTIL iclk_18_4 = '1';
    IF irst_n = '0' THEN
      lrck_1x_div  <=  (OTHERS => '0');
      lrck_1x    <=  '0';    
    ELSIF lrck_1x_div >= REF_CLK/(SAMPLE_RATE*2)-1 THEN
        lrck_1x_div  <=  (OTHERS => '0');
        lrck_1x  <=  NOT LRCK_1X;
      ELSE
        lrck_1x_div <= lrck_1x_div + 1;
    END IF;
  END PROCESS;  
  oAUD_LRCK <= LRCK_1X; -- Left/right control signal about 192 clock cycles

--------------------------------------------------
----------  16 Bits PISO MSB First  --------------
  P3: PROCESS
  BEGIN
    WAIT UNTIL oaud_bck = '0';
    IF irst_n = '0' THEN
      sel_cont  <=  (OTHERS => '0');
    ELSE
      sel_cont  <=  sel_cont + 1;
    END IF;
  END PROCESS;
  nsel_cont <= NOT ('1' & sel_cont); -- The transfer is MSB first
  oaud_data <= idata16(CONV_INTEGER(nsel_cont)); -- Single bit output
                       
END ARCHITECTURE FPGA;
